/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _SS_PATCHDELTA_H
#define _SS_PATCHDELTA_H

#include <stdint.h>
#include "sha1.h"
#include "unistd.h"
#include "fcntl.h"
#include "errno.h"
#include "SS_Engine_Update.h"

//#define ENHANCED_BSDIFF
#define SS_UPDATE_FS 0
#define SS_UPDATE_IMG 1
//#define SHA_DIGEST_SIZE 20
typedef struct {
	int type;
	ssize_t size;
	char *data;
} Value;
typedef struct _Patch {
	uint8_t sha1[SHA_DIGEST_SIZE];
	const char *patch_filename;
} Patch;

typedef struct _FileInfo {
	unsigned char sha1[20];     //SHA_DIGEST_SIZE 20
	unsigned char *data;
	int size;
	struct stat st;
} FileInfo;

typedef ssize_t(*SinkFn) (unsigned char *, ssize_t, void *);

int ParseSha1(const char *str, uint8_t * digest);

void ShowBSDiffLicense();
int ApplyBSDiffPatch(const unsigned char *old_data, ssize_t old_size,
		const Value * patch, ssize_t patch_offset, SinkFn sink, void *token, sha1_ctx_t * ctx1);
int ApplyBSDiffPatchMem(const unsigned char *old_data, ssize_t old_size,
		const Value * patch, ssize_t patch_offset, unsigned char **new_data, ssize_t * new_size);
//int ApplyOptimizedBSDiffPatch(const unsigned char* old_data,void* token,
//                              const Value* patch, SinkFn sink,sha1_ctx_t* ctx1);

int SS_LoadPartition(const char *filename, FileInfo * file);
int SS_LoadFile(const char *filename, FileInfo * file);
extern void SS_SetUpgradeState(int Val);
extern long SS_GetAvailableFreeSpace(const char *partition_name, SS_UINT32 * available_flash_size);
extern int SS_BackupSource(const char *source_filename);
extern int SS_ApplyBsdiff(char *oldfile, char *newfile, char *patch, SinkFn sink, void *token, sha1_ctx_t * ctx1);
extern int SS_BackupSourceClear();
extern int SS_PatchSourceClear();
extern long SS_WriteFile(long wHandle, SS_UINT32 dwPosition, unsigned char *pbBuffer, SS_UINT32 dwSize);
extern void SS_Free(void * pMemBlock);
extern long SS_CopyFile(const char * strFromPath, const char * strToPath);
extern long SS_DeleteFolder(const char * strPath);
extern long SS_DeleteFile(const char * strPath);
extern int tar_get_item_size(char * tar, char * item);
extern int tar_extract_file(char * tar, char * item, char * pathname);
extern int _system_cmd_wait(const char * command);
#endif
