/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*HEADER */

/*

Function Prototypes Mandatory

*/

#include<stdio.h>
#include<ctype.h>
#include<stdlib.h>
#include<string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/statfs.h>
#include <mntent.h>
#include "ua_types.h"
#include "SS_Common.h"
#include "fota_common.h"
#include "SS_UPI.h"
#include "SS_PatchDelta.h"
#include "SS_Engine_Errors.h"
#include "SS_FSUpdate.h"

int gtotalFSCnt = 0;
int FS_UpgradeState = E_SS_FAILURE;
int gvalid_session = 0;		 //used as fail-safe in case device is turmed off or battery removed during update
fs_list *fs_headptr_main = NULL;
fs_list *headptr_list[UA_PARTI_MAX];
tar_Data_t *tar_cfg_data = NULL;

#ifdef MEM_PROFILING
/*
	Description:
		Create script file , set it executable, execute script in child process
		Only works if valid delta.tar is present for upgrade
	Summary:
		If MEM_PROFILING is activated,
		we can see the result of memory profiling after delta upgrade
		in file defined by macro - SS_MEMORY_PROFILING_SCRIPT
*/
int mem_profiling_start = 0;
int SS_Do_Memory_Profiling()
{
	int ret = -1;
	pid_t pid;
	char memory_usage_script[1024] = "#!/bin/bash\nlog_file=$1\npid=$2\nmaxmem=0\nwhile [[ -d \"/proc/${pid}\" ]]; do\n\
					  mem=`cat /proc/${pid}/smaps | grep Pss | grep -v Swap|awk '{print $2}'|awk '{s+=$1} END {print s}'`\n\
					  if [[ ${mem} -gt ${maxmem} ]]; then\nmaxmem=${mem}\n\
						  echo -e \"Memory usage till now is: ${maxmem} KB.\" >> $log_file\n	fi\n	sleep 0.01\ndone\n\
							  echo -e \"Max was : ${maxmem} KB.\" >> $log_file\n";
	char cmd[1024] = { 0, };

	//Open a file and write the script contents in it
	FILE *fp = fopen(SS_MEMORY_PROFILING_SCRIPT, "w+");
	fwrite(memory_usage_script, strlen(memory_usage_script), 1, fp);
	fclose(fp);
	//make the file executable	  -	   Octal 495 is 757 decimal
	if (chmod(SS_MEMORY_PROFILING_SCRIPT, 495) < 0) {
		LOGE("Error in chmod(%s, 495) - %d (%s)\n", SS_MEMORY_PROFILING_SCRIPT, errno, strerror(errno));
		return E_SS_FAILURE;
	}
	//calling mem_use.sh
	//Usage : <location of script> <location of log file to be created> <pid of the calling process>
	pid = getpid();
	snprintf(cmd, sizeof(cmd) - 1, "%s %s %d", SS_MEMORY_PROFILING_SCRIPT, SS_MEMORY_USAGE_LOG, pid);
	ret = _system_cmd_nowait(cmd);
	sleep(1);
	LOG("ret for memory profiling cmd is %d\n", ret);
	if (ret == 0) {
		mem_profiling_start = 1;
		return S_SS_SUCCESS;
	} else {
		LOGE("Could not start Memory Profiling\n");
		return E_SS_FAILURE;
	}
}
#endif
static void hex_digest(char * sha1, char *buffer, int size)
{
	int i = 0;
	for ( i = 0; i < size; i++){
		snprintf(&buffer[i * 2], (size * 2) - (i * 2) + 1, "%02x", sha1[i]);
	}
}
#ifdef TIME_PROFILING
static char ts1[256];
static double ts2;
double t1;
double t2;
double fast_tar_get_item_size_time = 0.0;
double SS_LoadFile_time = 0.0;
double SS_FSBuildNodes_time = 0.0;

static void get_time_stamp1(void)
{
	struct timeval tv;
	int sec, msec;

	gettimeofday(&tv, NULL);
	sec = (int)tv.tv_sec;
	msec = (int)(tv.tv_usec / 1000);
	snprintf(ts1, 256, "%06d.%03d", sec % 100000, msec);
}

static double get_time_stamp2(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	ts2 = (double)tv.tv_sec + (double)(tv.tv_usec / 1000000.0);
	return ts2;
}
#endif

//Check SS function if available
int file_exist(char *filename)
{
	struct stat buf;
	int ret = 0;

	ret = lstat(filename, &buf);
	if (ret < 0)
		ret = stat(filename, &buf);
	return (ret >= 0) ? (1) : (0);
}

#ifdef POWER_FAIL_TEST

static int SS_Get_power_fail_flag(void)
{
	int fd;
	char buf[256];
	char *ptr = NULL;
	int result = 0;
	int ret_val = 0;

	if (file_exist(SS_POWER_FAIL_TEST_FLAG) == 0) {
		LOGE("No exist file!!\n");
		return -1;
	}

	fd = open(SS_POWER_FAIL_TEST_FLAG, O_RDWR, S_IRWXU);
	if (fd == -1) {
		LOGE("Could not open status file!!\n");
		return -1;
	}

	result = SS_ReadFile(fd, 0, buf, sizeof(buf));
	if (result != 0) {
		LOGE("SS_ReadFile failed!!\n");
		return -1;
	}

	ret_val = atoi(buf);

	result = SS_CloseFile(fd);
	if (result != 0)
		LOGE("SS_CloseFile failed!!\n");

	return ret_val;
}

static void SS_Set_power_fail_flag(int del_type)
{
	int fd;
	int result = 0;
	char num_str[16];
	LOG("del_type:[%d]\n", del_type);

	fd = open(SS_POWER_FAIL_TEST_FLAG, O_RDWR | O_CREAT, S_IRWXU);
	if (fd == -1) {
		LOGE("Could not open status file!!\n");
		return;
	}

	sprintf(num_str, "%d", del_type);
	result = SS_WriteFile(fd, 0, num_str, strlen(num_str));
	if (result != 0)
		LOGE("SS_WriteFile failed!!\n");

	result = SS_CloseFile(fd);
	if (result != 0)
		LOGE("SS_CloseFile failed!!\n");

	sync();

}

int SS_Do_Power_fail_test(int del_type)
{
	int ret = -1;
	char cmd[1024] = { 0, };

	SS_Set_power_fail_flag(del_type);

	snprintf(cmd, sizeof(cmd) - 1, "%s", "/usr/sbin/reboot fota");
	ret = _system_cmd_nowait(cmd);
	sleep(1);
	LOG("ret for SS_Do_Power_fail_test cmd is %d\n", ret);
	if (ret == 0) {
		return S_SS_SUCCESS;
	} else {
		LOGE("Could not start Memory Profiling\n");
		return E_SS_FAILURE;
	}
}
#endif

long SS_GetUPIVersion(unsigned char *ver_str)
{
	if (ver_str) {
		strncpy((char *)ver_str, SS_TOTA_VERSION, MAX_PATH);
#ifdef MEM_PROFILING
		if (!mem_profiling_start)
			if (!(S_SS_SUCCESS == SS_Do_Memory_Profiling()))
				LOGE("Unable to start Memory_Profiling");
#endif
		return S_SS_SUCCESS;//wgid: 2456
	} else
		return E_SS_FAILURE;
}

int SS_CalculateFileSha(char *filename, int filesize, FileInfo * file)
{

	FILE *fp = NULL;
	int ulResult = S_SS_SUCCESS;
	int chunk = 20*1024*1024;
	char buf[256];

	fp = fopen(filename, "rb");
	if (fp == NULL) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to open \"%s\": %s\n", filename, buf);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	file->data = SS_Malloc(chunk);
	if (!file->data) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to allocate memory for \"%s\": %s\n", filename, buf);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	ssize_t bytes_read = 0;
	sha1_ctx_t sha_ctx;
	sha1_init(&sha_ctx);

	while (filesize > 0) {
		if (filesize < chunk) {
			bytes_read = fread(file->data, 1, filesize, fp);
			if (bytes_read != filesize) {
				LOGE("short read of \"%s\" (%ld bytes of %ld)\n", filename, (long)bytes_read, (long)file->size);
				ulResult = E_SS_FAILURE;
				goto Cleanup;
			}
			sha1_update(&sha_ctx, file->data, filesize);
			break;
		} else {
			bytes_read = fread(file->data, 1, chunk, fp);
			if (bytes_read != chunk) {
				LOGE("short read of \"%s\" (%ld bytes of %ld)\n", filename, (long)bytes_read, (long)file->size);
				ulResult = E_SS_FAILURE;
				goto Cleanup;
			}
			sha1_update(&sha_ctx, file->data, chunk);
			filesize -= chunk;
		}
	}

	sha1_final(&sha_ctx, (uint32_t *) &file->sha1);

Cleanup:
	if (fp)
		fclose(fp);
	if (file->data)
		SS_Free(file->data);
	return ulResult;
}

int SS_verify_DELTA_image(char *filename)
{

	FileInfo file;
	FILE *fp = NULL;
	char line[SS_TOKEN_MAXLINE_LEN] = { '\0' };
	char *delta_size = NULL;
	char *signature = NULL;
	char *sha1trg = NULL;
	char *saveptr = NULL;
	uint8_t target_sha1[SHA_DIGEST_SIZE] = { 0, };
	char cmd[512] = { 0, };
	char buf[256];
	int udelta_size = 0;
	int ulResult = S_SS_SUCCESS;

	if (stat(filename, &file.st) != 0) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to stat \"%s\": %s\n", filename, buf);
		return -1;
	}

	snprintf(cmd, sizeof(cmd) - 1, "grep -o -P '%s' --binary-files=text system/%s > %s",
				SS_IMAGE_MAGIC_KEY, filename, SS_IMAGE_MAGIC_KEY_VAL);

	ulResult = _system_cmd_wait(cmd);
	if (ulResult != S_SS_SUCCESS) {
		LOGE("Grep extraction for [%s] failed, code [%d]\n", cmd, ulResult);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	fp = fopen(SS_IMAGE_MAGIC_KEY_VAL, "r");
	if (!fp) {
		LOGE("Grep extraction for [%s] failed, code [%d]\n", cmd, ulResult);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	if (fgets(line, SS_TOKEN_MAXLINE_LEN, fp) == NULL) {
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}
	fclose(fp);
	fp = NULL;

	signature = strtok_r(line, SS_TOEKN_COLON, &saveptr);
	delta_size = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
	sha1trg = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);

	if (signature && sha1trg && delta_size) {
		udelta_size = atoi(delta_size);
	if (udelta_size < 0)
				LOGE("Invalid udelta_size %d (%s)\n", udelta_size, delta_size);
		LOGL(LOG_SSENGINE, "delta_size %d sha1trg %s\n", udelta_size, sha1trg);
	} else {
		LOGE("Could not parse signature [%s]\n", line);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	if (ParseSha1(sha1trg, target_sha1) != 0) {
		LOGE("failed to parse tgt-sha1 \"%s\"\n", sha1trg);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	ulResult = SS_CalculateFileSha(filename, udelta_size, &file);
	if (ulResult != S_SS_SUCCESS)
		goto Cleanup;

	if (memcmp(file.sha1, target_sha1, SHA_DIGEST_SIZE) != 0) {
		unsigned char actualShaBuffer[41] = { 0, };
		hex_digest(file.sha1, actualShaBuffer, SHA_DIGEST_SIZE);
		LOGE("SHA mismatch -[%s] actual [%s] target [%s]\n", filename, actualShaBuffer, sha1trg);
		SS_SetUpgradeState(E_SS_FSBADDELTA);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	} else {
		LOGL(LOG_SSENGINE, "DELTA verified %s \n", sha1trg);
	}

Cleanup:
	if (fp)
		fclose(fp);
	if (file_exist(SS_IMAGE_MAGIC_KEY_VAL))
		SS_DeleteFile(SS_IMAGE_MAGIC_KEY_VAL);
	return ulResult;
}


int SS_GetProgressResolution(int ultotalFSCnt)
{
	if (ultotalFSCnt < DISPLAYRESOLUTION_SIZE)
		return 1;
	else
		return (ultotalFSCnt / DISPLAYRESOLUTION_SIZE);
}

void SS_SetUpgradeState(int Val)
{
	LOGE("FAILED to upgrade Cause:[0x%x]\n", Val);
	FS_UpgradeState = Val;
	return;
}

int SS_GetUpgradeState()
{
	return FS_UpgradeState;
}

int SS_Get_last_update_status(int* last_update_status, int* del_type)
{
	int fd;
	unsigned char buf[256];
	char *ptr = NULL;
	char *saveptr = NULL;
	int result = 0;

	if (file_exist(SS_UPDATE_STATUR_PATH) == 0) {
		LOG("No exist file!! - File_path:[%s]\n", SS_UPDATE_STATUR_PATH);
		return -1;
	}

	fd = open(SS_UPDATE_STATUR_PATH, O_RDWR, S_IRWXU);
	if (fd == -1) {
		LOGE("Could not open status file!!, File_path:[%s]\n", SS_UPDATE_STATUR_PATH);
		return -1;
	}

	result = SS_ReadFile(fd, 0, buf, sizeof(buf));
	if (result != 0) {
		LOGE("SS_ReadFile failed!!\n");
		result = SS_CloseFile(fd);
		if (result != 0)
			LOGE("SS_CloseFile failed!!\n");
		return -1;
	}

	ptr = strtok_r((char *)buf, " ", &saveptr);

	if (ptr != NULL) {
		*last_update_status = atoi(ptr);
		ptr = strtok_r(NULL, " ", &saveptr);
	}

	if (ptr != NULL) {
		*del_type = atoi(ptr);
	}

	result = SS_CloseFile(fd);
	if (result != 0)
		LOGE("SS_CloseFile failed!!\n");

	return 0;
}

void SS_Set_last_update_status(int last_update_status, int del_type)
{
	int fd;
	int result = 0;
	char num_str[16];
	LOG("last_update_status:[%d], del_type:[%d]\n", last_update_status, del_type);

	fd = open(SS_UPDATE_STATUR_PATH, O_RDWR | O_CREAT, S_IRWXU);
	if (fd == -1) {
		LOGE("Could not open status file!!, File_path:[%s]\n", SS_UPDATE_STATUR_PATH);
		return;
	}

	snprintf(num_str, sizeof(num_str), "%d %d", last_update_status, del_type);
	result = SS_WriteFile(fd, 0, (unsigned char *)num_str, strlen(num_str));
	if (result != 0)
		LOGE("SS_WriteFile failed!!\n");

	result = SS_CloseFile(fd);
	if (result != 0)
		LOGE("SS_CloseFile failed!!\n");

	sync();

}

int SS_rename(const char *old_file_name, const char *new_file_name)
{
	if (link(old_file_name, new_file_name) < 0) {
		if (errno == EEXIST) {
			if (unlink(new_file_name) < 0 || link(old_file_name, new_file_name) < 0)
				return -1;
		} else
			return -1;
	}
	if (unlink(old_file_name) < 0) {
		if (unlink(new_file_name) == 0)
			return -1;
	}
	return 0;
}

int SS_rename1(const char *old_file_name, const char *new_file_name)
{

	int result = E_SS_FAILURE;
	char *temp_name = NULL;
	temp_name = (char *)SS_Malloc(strlen(new_file_name) + 10);
	if (temp_name == NULL)
		return E_SS_FAILURE;
	snprintf(temp_name, strlen(new_file_name) + 10, "%s.temp", new_file_name);
	result = rename(new_file_name, temp_name);
	if (result != 0)
		goto Cleanup;
	result = rename(old_file_name, new_file_name);
	if (result != 0)
		goto Cleanup;
	result = unlink(temp_name);
 Cleanup:
	if (temp_name)
		SS_Free(temp_name);
	return result;
}

/*!
 *********************************************************************************
 *					 SS_FSVerifyNode
 *********************************************************************************
 *
 * @brief
 *	This is to verify nodes being added to global control structure for diff and delete cases.
 *	gvalid_session global is used to check if nodes are already verified.
 *
 *
 *	@param
 *
 *	@return				returns S_SS_SUCCESS and
 *						returns E_SS_FAILURE in case any error occurs
 *
 *********************************************************************************
 */
int SS_FSVerifyNode(const char *path, const char *patchname, const char *sha1src, int type,
					char *patchpath_name, int *data_size, int *data_offset)
{
	char patch[MAX_FILE_PATH] = { 0 };
	FileInfo source_file;
	uint8_t source_sha1[SHA_DIGEST_SIZE];

	if (gvalid_session) {
		if ((type == DIFFS || type == DELETES || type == MOVES) && !file_exist((char *)path)) {
			LOGE("failed to verifyNodes [does not exist], Path : [%s] Type[%d]\n", path, type);
			SS_SetUpgradeState(E_SS_FSBADNODES);
			return E_SS_FAILURE;
		}
		snprintf(patch, MAX_FILE_PATH, "%s%s%s", patchpath_name, "/", patchname);
		if ((type == DIFFS/*||type == NEWFILES */)) {   // allowing size 0 also for folders
			if (tar_get_item_size_from_struct(&tar_cfg_data, patchname, data_size, data_offset)
					!= S_SS_SUCCESS) {
				LOGE("failed to get item size from struct, Patch : [%s]\n", patchname);
				SS_SetUpgradeState(E_SS_FSBADNODES);
				return E_SS_FAILURE;
			}
			if (*data_size < 0) {
				LOGE("failed to verifyNodes [delta absent], Patch : [%s] Type[%d]\n", patch, type);
				SS_SetUpgradeState(E_SS_FSBADNODES);
				return E_SS_FAILURE;
			}
		}
		//For other types (NEWs, SYMs, Folders), SHA check not required
		if ((type == DIFFS || type == MOVES/* ||type == DELETES */) &&
				(strcmp(sha1src, SS_NULLENTRY) != 0)) {
			if (ParseSha1(sha1src, source_sha1) != 0) {
				LOGE("Failed to parse Src-sha1 \"%s\"\n", sha1src);
				return E_SS_FAILURE;
			}
			if (SS_LoadFile(path, &source_file) == 0) {
				if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) != 0) {
					SS_Free(source_file.data);
					unsigned char actualShaBuffer[41] = { 0, };
					hex_digest(source_file.sha1, actualShaBuffer, SHA_DIGEST_SIZE);
					LOGE("SS_FSVerifyNode - SHA mismatch with SRC  - PATH [%s] Expected [%s] Actual [%s]\n",
						path, sha1src, actualShaBuffer);
					SS_SetUpgradeState(E_SS_FSSRCCURRUPTED);	// Define other error
					return E_SS_FAILURE;
				}
			}
			SS_Free(source_file.data);
		}
	} else {
		LOGL(LOG_SSENGINE, "FS partition Already verified - Filling only size and offset \n");
		snprintf(patch, MAX_FILE_PATH, "%s%s%s", patchpath_name, "/", patchname);
		if ((type == DIFFS/* ||type == NEWFILES */)) {   // allowing size 0 also for folders
			if (tar_get_item_size_from_struct(&tar_cfg_data, patchname, data_size, data_offset)
					!= S_SS_SUCCESS) {
				LOGE("failed to get item size from struct, Patch : [%s] \n", patchname);
				SS_SetUpgradeState(E_SS_FSBADNODES);
				return E_SS_FAILURE;
			}
			if (*data_size < 0) {
				LOGE("failed to verifyNodes [delta absent], Patch : [%s] Type[%d]\n", patch, type);
				SS_SetUpgradeState(E_SS_FSBADNODES);
				return E_SS_FAILURE;
			}
		}
	}
	return S_SS_SUCCESS;
}

/*!
 *********************************************************************************
 *					 SS_AppendNode
 *********************************************************************************
 *
 * @brief
 *	This is to append node to the global control structure for delta files.
 *
 *
 *	@param
 *
 *	@return				returns S_SS_SUCCESS and
 *						returns E_SS_FAILURE in case any error occurs
 *
 *********************************************************************************
 */
int SS_AppendNode(const char *ubDeltaPath, fs_params ** headparam, fs_params ** tailparam, const char *old_path,
				  const char *new_path, const char *patchname, const char *sha1src, const char *sha1trg, int type,
				  char *patchpath_name)
{
	fs_params *newnode = NULL;
	int data_size = 0;
	int data_offset = 0;

	if (!ubDeltaPath || !old_path || !new_path || !patchname || !sha1src || !sha1trg || !patchpath_name) {
		LOGE("Bad Nodes, NULL params passed for Appending Nodes \n");
		SS_SetUpgradeState(E_SS_FSINVALIDNODEPARAMS);
		return E_SS_FAILURE;
	}
	if ((E_SS_FAILURE ==
		 SS_FSVerifyNode(old_path, patchname, sha1src, type, patchpath_name, &data_size, &data_offset))) {
		LOGE("Bad Nodes, Failed to pass verification - [Delta Path - %s][OldPath - %s] [NewPath - %s] \n", ubDeltaPath,
			 old_path, new_path);
		return E_SS_FAILURE;
	}
	newnode = (fs_params *) SS_Malloc(sizeof(fs_params));
	if (!newnode)
		return E_SS_FAILURE;
	strncpy(newnode->file_old_path, old_path, SS_MAX_NAMELENSUPPORTED);//wgid: 29483
	strncpy(newnode->file_new_path, new_path, SS_MAX_NAMELENSUPPORTED);//wgid: 29482
	strncpy(newnode->patch_name, patchname, SS_MAX_NAMELENSUPPORTED);//wgid: 28033
	strncpy(newnode->sha1src, sha1src, sizeof(newnode->sha1src) -1);//wgid: 25282
	strncpy(newnode->sha1trg, sha1trg, sizeof(newnode->sha1trg) - 1);//wgid: 25283
	newnode->type = type;
	newnode->data_size = data_size;
	newnode->data_offset = data_offset;
	newnode->nextnode = NULL;

	//LOG("%s %s %d %s %s \n",newnode->file_path,newnode->patch_name,newnode->type, newnode->sha1src, newnode->sha1trg);

	if (*headparam == NULL) {
		*headparam = newnode;
		*tailparam = newnode;
	} else {
		(*tailparam)->nextnode = newnode;
		(*tailparam) = (*tailparam)->nextnode;
	}
	return S_SS_SUCCESS;

}

void SS_UpdateUIProgress(ua_dataSS_t * ua_dataSS, int ulTotalFsCnt, int ulDone)
{
	static int ss_count = 1;
	int res_val = SS_GetProgressResolution(ulTotalFsCnt);
	if (!ua_dataSS) {
		LOGE("Error ua_dataSS\n");
		return;
	}
//LOG("\nvalues are ss_count[%d] total_file_cnt[%d]",ss_count,ulTotalFsCnt);
	if (ulDone == 1) {
		if (ua_dataSS->ui_progress)
			ua_dataSS->ui_progress(ua_dataSS, 100);
		ss_count = 1;
	} else if (ss_count < ulTotalFsCnt) {
		if (ss_count % res_val == 0) { //Max 50 times display
			double data = (double)ss_count / (double)ulTotalFsCnt;
			if (ua_dataSS->ui_progress)
				ua_dataSS->ui_progress(ua_dataSS, data * 100);
		}
		ss_count++;
	} else {
		if (ua_dataSS->ui_progress)
			ua_dataSS->ui_progress(ua_dataSS, 100);
		ss_count = 1;
	}

}

/*!
 *********************************************************************************
 *					 SS_FSClearNodes
 *********************************************************************************
 *
 * @brief
 *	This is to clear the global control structure for delta files.
 *
 *
 *	@param
 *
 *	@return
 *
 *********************************************************************************
 */
void SS_FSClearNodes(int idx)
{
	fs_params *local_temp = NULL;
	fs_params *local_next = NULL;
	LOGL(LOG_SSENGINE, "Free Nodes idx=%d \n", idx);
	if (headptr_list[idx]) {
		if (headptr_list[idx]->del_ref) {
			local_temp = headptr_list[idx]->del_ref;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		if (headptr_list[idx]->dif_ref) {
			local_temp = headptr_list[idx]->dif_ref;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		if (headptr_list[idx]->move_ref) {
			local_temp = headptr_list[idx]->move_ref;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		if (headptr_list[idx]->new_ref) {
			local_temp = headptr_list[idx]->new_ref;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		if (headptr_list[idx]->sym_difref) {
			local_temp = headptr_list[idx]->sym_difref;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		if (headptr_list[idx]->sym_newref) {
			local_temp = headptr_list[idx]->sym_newref;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		SS_Free(headptr_list[idx]);
		headptr_list[idx] = NULL;
	}
}

/*!
 *********************************************************************************
 *					 SS_FSGetDeltaCount
 *********************************************************************************
 *
 * @brief
 *	This is to get the delta count for diffs , deletes etc.
 *
 *
 *	@param
 *
 *	@return				returns structure with delta count info
 *						NULL in case of error
 *
 *********************************************************************************
 */

struct details *SS_FSGetDeltaCount(char *ubDeltaPath, char *ubDeltaInfoFile)
{
	int size = 0, bckupsize = 0, ret = 1;
	char *token = NULL;
	char *FileData = NULL;
	int data_size = -1;
	char *line = NULL;
	char *saveptr = NULL;
	char buf[256];
	struct details *refer_copy = NULL;
	FILE *filename_bkup = NULL;

	if (!(ubDeltaPath && ubDeltaInfoFile)) {
		LOGE("failed to Parse DELTA count information: \n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return NULL;
	}
	refer_copy = (struct details *)SS_Malloc(sizeof(struct details));

	if (refer_copy == NULL) {
		LOGE("failed to allocate memory\n");
		return NULL;
	}

	LOGL(LOG_SSENGINE, "Delta File Info =%s\n", ubDeltaInfoFile);

	size = tar_get_item_size(ubDeltaPath, ubDeltaInfoFile);
	if (size < 0) {
		LOGE("failed to Access DELTA info file DPath:[%s] File: [%s]\n", ubDeltaPath, ubDeltaInfoFile);
		SS_SetUpgradeState(E_SS_FSBADDELTA);
		ret = 0;
		goto cleanup;
	}

	FileData = SS_Malloc(size + 1);
	if (FileData == NULL) {
		LOGE("Failed to Allocate Memory\n");
		SS_SetUpgradeState(E_SS_MALLOC_ERROR);
		ret = 0;
		goto cleanup;
	}

	memset(FileData, 0, size + 1);
	data_size = tar_get_cfg_data(ubDeltaPath, ubDeltaInfoFile, FileData, size);
	if (data_size <= 0) {	   // == 0 is NOT okay??
		LOGE("Failed to read cfg data from Delta\n");
		SS_SetUpgradeState(E_SS_FSBADDELTA);
		ret = 0;
		goto cleanup;
	}
	filename_bkup = fopen(SS_PATCHLIST_BKUPLOC, "wb+");
	if (filename_bkup == NULL) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("Failed to create BACKUP file Error:[%s]\n", buf);
		SS_SetUpgradeState(E_SS_FSFAILEDTOBACKUPPATCHINFO);
		ret = 0;
		goto cleanup;
	}

	bckupsize = fwrite(FileData, 1, data_size, filename_bkup);  //RECHECK SIZE 1
	if (bckupsize <= 0) {
		SS_SetUpgradeState(E_SS_FSFAILEDTOBACKUPPATCHINFO);
		ret = 0;
		goto cleanup;
	}
	LOGL(LOG_SSENGINE, " Size [%d] DataSize [%d] BakUpSize [%d]\n", size, data_size, bckupsize);

	line = strstr(FileData, SS_FSCOUNT_MAGIC_KEY);
	if (line) {
		LOGL(LOG_SSENGINE, "SS_FSGetDeltaCount() last line %s \n", line);

		token = strtok_r(&line[SS_FSCOUNT_MAGIG_KEYLEN], SS_TOKEN_SPACE, &saveptr);
		if (token)
			refer_copy->diffs = atoi(token);
		else {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
			ret = 0;
			goto cleanup;
		}
		token = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr);
		if (token)
			refer_copy->moves = atoi(token);
		else {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
			ret = 0;
			goto cleanup;
		}

		token = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr);
		if (token)
			refer_copy->news = atoi(token);
		else {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
			ret = 0;
			goto cleanup;
		}

		token = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr);
		if (token)
			refer_copy->deletes = atoi(token);
		else {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
			ret = 0;
			goto cleanup;
		}

		token = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr);
		if (token)
			refer_copy->symdiffs = atoi(token);
		else {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
			ret = 0;
			goto cleanup;
		}

		token = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr);
		if (token)
			refer_copy->symnews = atoi(token);
		else {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
			ret = 0;
			goto cleanup;
		}

		gtotalFSCnt =
			refer_copy->diffs + refer_copy->moves + refer_copy->news + refer_copy->deletes + refer_copy->symdiffs +
			refer_copy->symnews;
		LOG("SS_FSGetDeltaCount() total no of file %d\n", gtotalFSCnt);

	} else {
		SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTACNT);
		LOG("SS_FSGetDeltaCount() Failed to read last line\n");
	}
	if (gtotalFSCnt < 0) {
		ret = 0;
		goto cleanup;
	}

 cleanup:
	if (ret) {
		SS_Free(FileData);
		if (filename_bkup)
			fclose(filename_bkup);
		return refer_copy;
	} else {
		SS_Free(FileData);
		SS_Free(refer_copy);
		if (filename_bkup)
			fclose(filename_bkup);
		return NULL;
	}

}

/*!
 *********************************************************************************
 *					 SS_FSBuildNodes
 *********************************************************************************
 *
 * @brief
 *	This is used to build the gobal control structure for diffs, deletes etc.
 *	For all the entries in config file (which has info) the information is parsed to the global control struct
 *
 *
 *	@param
 *
 *	@return				returns fs_list structure filled with details of all the files to be diff-ed ,deleted etc.
 *						NULL in case of error
 *
 *********************************************************************************
 */
fs_list *SS_FSBuildNodes(ua_dataSS_t * ua_dataSS)
{
	FILE *fp = NULL;
	char line[SS_TOKEN_MAXLINE_LEN] = { '\0' };
	char string_na[] = "NA";
	char *patch_name = NULL;
	char *source_name = NULL;
	char *target_name = NULL;
	char *sha1src = NULL;
	char *sha1trg = NULL;
	char *change_type = NULL;
	char *file_type = NULL;
	char *saveptr = NULL;
	uint32_t ulPatchCount = 0, del_type = DELETES;
	fs_params *fs_diffhead = NULL;
	fs_params *fs_difftail = NULL;
	fs_params *fs_movehead = NULL;
	fs_params *fs_movetail = NULL;
	fs_params *fs_newhead = NULL;
	fs_params *fs_delhead = NULL;
	fs_params *fs_deltail = NULL;
	fs_params *fs_symlinkdiffhead = NULL;
	fs_params *fs_symlinkdifftail = NULL;
	fs_params *fs_symlinknewhead = NULL;
	fs_params *fs_symlinknewtail = NULL;

	struct details *local = NULL;
	fs_list *fs_head_node = NULL;
	int i = 0, retval = 0;
	if (!ua_dataSS) {
		LOGE("Bad structure ua_dataSS\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return NULL;
	}
	LOGL(LOG_SSENGINE, " Build Nodes Entry \n");
	if (tar_cfg_data == NULL)
		tar_cfg_data = tar_build_cfg_table(ua_dataSS->update_data->ua_delta_path);
	if (!tar_cfg_data) {
		LOGE(" tar_build_cfg_table  Failed \n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return NULL;
	}
	local = SS_FSGetDeltaCount(ua_dataSS->update_data->ua_delta_path, ua_dataSS->update_delta->ua_patch_info);
	if (local == NULL) {
		LOGE(" Build Nodes Failed \n");
		if (tar_cfg_data)
			tar_free_cfg_table(&tar_cfg_data);
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return NULL;
	}

	fp = fopen(SS_PATCHLIST_BKUPLOC, "r");
	if (!fp) {
		SS_SetUpgradeState(E_SS_FSFAILEDTOOPENPATCHINFO);
		if (tar_cfg_data)
			tar_free_cfg_table(&tar_cfg_data);
		SS_Free(local);
		return NULL;
	}

	ulPatchCount = local->diffs + local->deletes + local->news + local->moves + local->symdiffs + local->symnews;
	LOG("Total FS count [%d].\n", ulPatchCount);
/*
************************************************************************
Parsing logic implemented for patchlist
************************************************************************
Sample entries in patchlist as below :
<CHANGE-TYPE>:<FILE-TYPE>:<OTHER-DETAILS...>
************************************************************************
DIFF:REG:system/bin/vi:system/bin/vi:2f2f3dc6d3ee06af0080ac7975f22941660f2480:78b2d44af32d854c70f1cb7431a60c2682a320cc:diff1_vi.delta
DIFF:TPK:system/usr/packages/removable/com.samsung.calculator.tpk:system/usr/packages/removable/com.samsung.calculator.tpk:
				 96fc1bcde30d501ba65ef0038e05da46d255a7b3:fa1d5d9daa4097ac302b69244297f508577c3a01:diff1598_com.samsung.calculator.tpk.delta/
MOVE:REG:system/etc/smack/accesses.d/heremaps-engine-devel:system/usr/apps/com.samsung.contacts/res/temp:da39a3ee5e6b4b0d3255bfef95601890afd80709
DEL:REG:system/usr/ug/res/images/ug-phone/contacts/favorites_icon_remove.PNG:38ad8be378506d19b1c769d46be262cf100f6c59
DEL:SYM:system/usr/apps/com.samsung.message-lite/lib/libmsg-common.so
SYM:DIFF:system/usr/lib/sync-agent/kies-private/libplugin-na-mobex.so.0:system/usr/lib/sync-agent/kies-private/libplugin-na-mobex.so.0:
					libplugin-na-mobex.so.0.3.57
SYM:NEW:system/lib/firmware/vbc_eq:/opt/system/vbc_eq
***********************************************************************
*/
	if (local && ((local->diffs) > 0 || (local->moves > 0))) {
		LOGL(LOG_SSENGINE, "%ss [%d] %ss [%d]\n", SS_STRING_DIFF, local->diffs, SS_STRING_MOVE, local->moves);
		for (i = 0; i < (local->diffs + local->moves); i++) {
			if (fgets(line, SS_TOKEN_MAXLINE_LEN, fp) == NULL) {
				SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
				break;
			}
			//LOGL(LOG_SSENGINE, "DIFF LINE:[%d] [%s] \n",i+1,line);

			change_type = strtok_r(line, SS_TOEKN_COLON, &saveptr);
			file_type = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);

			if (change_type && strcmp(change_type, SS_STRING_MOVE) == 0) {	 // && strcmp(file_type,"TPK") == 0){
				source_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				target_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				sha1src = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				//LOGL(LOG_SSENGINE, "%s Index [%d]\n", SS_STRING_MOVE, i);

				if (!source_name || !target_name || !sha1src) {
					SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
					//LOGE("Failed to extract Patch Info Type:DELETES \n");
					LOGE("Failed to parse DIFFS - LINE:[%d] [%s] \n", i + 1, line);
					goto CleanUp;
				}
				retval =
					SS_AppendNode(ua_dataSS->update_data->ua_delta_path, &fs_movehead, &fs_movetail, source_name,
								  target_name, string_na, sha1src, string_na, MOVES,
								  ua_dataSS->update_delta->ua_patch_path);
				if (retval == E_SS_FAILURE)	 // ONLY test purpose, should enable this
					goto CleanUp;
		if (ua_dataSS->ua_operation == UI_OP_SCOUT) {
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
		}
			} else if (change_type && strcmp(change_type, SS_STRING_DIFF) == 0) {	  // && strcmp(file_type,"TPK") == 0){
				source_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				target_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				sha1src = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				sha1trg = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				patch_name = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				//LOGL(LOG_SSENGINE, "%s Index [%d]\n", SS_STRING_DIFF, i);

				if (patch_name && (strlen(patch_name) <= SS_MAX_NAMELENSUPPORTED)) {
					retval =
						SS_AppendNode(ua_dataSS->update_data->ua_delta_path, &fs_diffhead, &fs_difftail,
									  source_name, target_name, patch_name, sha1src, sha1trg, DIFFS,
									  ua_dataSS->update_delta->ua_patch_path);
					if (retval == E_SS_FAILURE) // ONLY test purpose, should enable this
						goto CleanUp;
				} else {
					SS_SetUpgradeState(E_SS_FILENAMELENERROR);
					LOGE("File Name length Limitation Error File:[%s]\n", patch_name);
					goto CleanUp;
				}
		if (ua_dataSS->ua_operation == UI_OP_SCOUT) {
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
		}
			} else {
				SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
				LOGE("Patch Name format Error File\n");
				goto CleanUp;
			}
		}
	}
	if (local && (local->news) > 0) {   //check if new files archive is present in the delta
		char new_patch_path[MAX_FILE_PATH] = { 0, };
		snprintf(new_patch_path, MAX_FILE_PATH, "%s%s", ua_dataSS->parti_info->ua_subject_name, SS_COMPRESSED_FILE);
		if (tar_get_item_size(ua_dataSS->update_data->ua_delta_path, new_patch_path) <= 0) {
			SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
			LOGE("New files not present in Patch\n");
			goto CleanUp;
		}
	}
	if (local && (local->deletes) > 0) {		//this is to group to delete list
		LOGL(LOG_SSENGINE, "%ss [%d]\n", SS_STRING_DEL, local->deletes);
		for (i = 0; i < (local->deletes); i++) {
			if (fgets(line, SS_TOKEN_MAXLINE_LEN, fp) == NULL) {
				SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
				break;
			}

			change_type = strtok_r(line, SS_TOEKN_COLON, &saveptr);
			file_type = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);

			if (file_type && strcmp(file_type, SS_STRING_REG) == 0) {
				source_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				sha1src = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				del_type = DELETES;
			} else if (file_type && strcmp(file_type, SS_STRING_SYM) == 0) {
				source_name = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				sha1src = string_na;
				del_type = DELETES;
			} else if (file_type && strcmp(file_type, SS_STRING_END) == 0) {
				source_name = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				sha1src = string_na;
				del_type = DELETE_END;
			} else {
				LOGE("Failed to parse DELETES - LINE:[%d] [%s] \n", i + 1, line);
				goto CleanUp;
			}

			if (!source_name || !sha1src) {
				SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
				//LOGE("Failed to extract Patch Info Type:DELETES \n");
				LOGE("Failed to parse DELETES - LINE:[%d] [%s] \n", i + 1, line);
				goto CleanUp;
			}
			//LOGL(LOG_SSENGINE, "%s Index [%d]\n", SS_STRING_DEL, i);
			retval =
				SS_AppendNode(ua_dataSS->update_data->ua_delta_path, &fs_delhead, &fs_deltail, source_name,
							  string_na, string_na, sha1src, string_na, del_type,
							  ua_dataSS->update_delta->ua_patch_path);
			if (retval == E_SS_FAILURE) // ONLY test purpose, should enable this
				goto CleanUp;
	    if (ua_dataSS->ua_operation == UI_OP_SCOUT) {
			SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
	    }

		}
	}						   //For symlink files

	if (local && (local->symdiffs) > 0) {
		LOGL(LOG_SSENGINE, "%s %ss [%d]\n", SS_STRING_SYM, SS_STRING_DIFF, local->symdiffs);
		for (i = 0; i < (local->symdiffs); i++) {	   //get the count from below function
			if (fgets(line, SS_TOKEN_MAXLINE_LEN, fp) == NULL) {
				SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
				break;
			}
			//LOGL(LOG_SSENGINE, "SYMDIFF LINE:[%d] [%s] \n",i+1,line);

			change_type = strtok_r(line, SS_TOEKN_COLON, &saveptr);
			file_type = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);

			if ((change_type && file_type) &&
					strcmp(change_type, SS_STRING_SYM) == 0 && strcmp(file_type, SS_STRING_DIFF) == 0) {	// && strcmp(file_type,"TPK") == 0){
				source_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				target_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				patch_name = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				//LOGL(LOG_SSENGINE, "%s %s Index [%d]\n", SS_STRING_SYM, SS_STRING_DIFF, i);

				if (!source_name || !target_name || !patch_name) {
					SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
					//LOGE("Failed to extract Patch Info Type:DELETES \n");
					LOGE("Failed to parse SymDiffs - LINE:[%d] [%s] \n", i + 1, line);
					goto CleanUp;
				}
				retval =
					SS_AppendNode(ua_dataSS->update_data->ua_delta_path, &fs_symlinkdiffhead, &fs_symlinkdifftail,
								  source_name, target_name, patch_name, string_na, string_na, SYMDIFFS,
								  ua_dataSS->update_delta->ua_patch_path);
				if (retval == E_SS_FAILURE)	 // ONLY test purpose, should enable this
					goto CleanUp;
		if (ua_dataSS->ua_operation == UI_OP_SCOUT) {
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
		}
			}
		}
	}
	if (local && (local->symnews) > 0) {
		LOGL(LOG_SSENGINE, "%s %ss [%d]n", SS_STRING_SYM, SS_STRING_NEW, local->symnews);
		for (i = 0; i < (local->symnews); i++) {
			if (fgets(line, SS_TOKEN_MAXLINE_LEN, fp) == NULL) {
				SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
				break;
			}
			//LOGL(LOG_SSENGINE, "SYMNEWS LINE:[%d] [%s] \n",i+1,line);

			change_type = strtok_r(line, SS_TOEKN_COLON, &saveptr);
			file_type = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);

			if ((change_type && file_type) &&
					(strcmp(change_type, SS_STRING_SYM) == 0 && strcmp(file_type, SS_STRING_NEW) == 0)) {
				source_name = strtok_r(NULL, SS_TOEKN_COLON, &saveptr);
				patch_name = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr);
				//LOGL(LOG_SSENGINE, "%s %s Index [%d]\n", SS_STRING_SYM, SS_STRING_NEW, i);

				if (!source_name || !patch_name) {
					SS_SetUpgradeState(E_SS_FSFAILEDTOPARSEDELTAINFO);
					//LOGE("Failed to extract Patch Info Type:DELETES \n");
					LOGE("Failed to parse SymNews - LINE:[%d] [%s] \n", i + 1, line);
					goto CleanUp;
				}
				retval =
					SS_AppendNode(ua_dataSS->update_data->ua_delta_path, &fs_symlinknewhead, &fs_symlinknewtail,
								  source_name, string_na, patch_name, string_na, string_na, SYMNEWFILES,
								  ua_dataSS->update_delta->ua_patch_path);
				if (retval == E_SS_FAILURE)	 // ONLY test purpose, should enable this
					goto CleanUp;
		if (ua_dataSS->ua_operation == UI_OP_SCOUT) {
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
		}
			}
		}
	}

	fs_head_node = (fs_list *) SS_Malloc(sizeof(fs_list));
	if (!fs_head_node) {
		SS_SetUpgradeState(E_SS_MALLOC_ERROR);
		goto CleanUp;
	}
	fs_head_node->dif_ref = fs_diffhead;
	fs_head_node->move_ref = fs_movehead;
	fs_head_node->new_ref = fs_newhead;
	fs_head_node->del_ref = fs_delhead;
	fs_head_node->sym_difref = fs_symlinkdiffhead;
	fs_head_node->sym_newref = fs_symlinknewhead;
	fs_head_node->ulPatchCount = ulPatchCount;

	if (ua_dataSS->ua_operation == UI_OP_SCOUT) {
	SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 1);
	}

 CleanUp:
	fclose(fp);
	SS_Free(local);
	unlink(SS_PATCHLIST_BKUPLOC);
	if (retval == E_SS_FAILURE)
		if (tar_cfg_data)
			tar_free_cfg_table(&tar_cfg_data);
	return fs_head_node;
}

void SS_GetPartition_LocDetails(ua_dataSS_t * ua_dataSS)
{
	LOGL(LOG_SSENGINE, "PART NAME: [%s] \n", ua_dataSS->parti_info->ua_parti_name);
	snprintf(ua_dataSS->update_delta->ua_patch_path, MAX_FILE_PATH, "%s", ua_dataSS->parti_info->ua_subject_name);
	snprintf(ua_dataSS->update_delta->ua_patch_info, MAX_FILE_PATH, "%s%s%s", ua_dataSS->parti_info->ua_subject_name,
			 ua_dataSS->parti_info->ua_parti_name, SS_PATCHLISTFORMAT);
	snprintf(ua_dataSS->update_delta->ua_attrib_path, MAX_FILE_PATH, "%s%s%s", ua_dataSS->parti_info->ua_subject_name,
			 ua_dataSS->parti_info->ua_parti_name, SS_PATCH_ATTR_FORMAT);
	LOGL(LOG_SSENGINE, "PatchPath[%s] PatchInfo [%s] Attributes [%s]\n", ua_dataSS->update_delta->ua_patch_path,
		 ua_dataSS->update_delta->ua_patch_info, ua_dataSS->update_delta->ua_attrib_path);

	return;
}

//Support functions//Change Struct format details (Can include total file count also???)/*!
/*
******************************************************************************** *
SS_FSSetAttributes
	*********************************************************************************
	* *@brief
	* This is used to set the file attributes at the end of application of patches in FS
	*
	* *@param
	*
	*@return returns S_SS_SUCCESS
	* E_SS_FAILURE in case of error
	*
	*********************************************************************************
 */
int SS_FSSetAttributes(ua_dataSS_t * ua_dataSS)
{
	char *pline = NULL;
	char *psaveptr = NULL;
	char *pfilePath = NULL;
	char *pfiletype = NULL;
	char *attributSize = NULL;
	char *pattribs = NULL;
	int ulAttribSize = 0;
	int result = S_SS_SUCCESS;
	int fail_cnt = 0;

	if (!(ua_dataSS && ua_dataSS->update_delta && ua_dataSS->update_data->ua_delta_path)) {
		LOGE("Bad params for SS_FSSetAttributes\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}
	LOGL(LOG_SSENGINE, "ATTRIB PATH: [%s] \n", ua_dataSS->update_delta->ua_attrib_path);

	int item_size = tar_get_item_size(ua_dataSS->update_data->ua_delta_path, ua_dataSS->update_delta->ua_attrib_path);

	if (item_size <= 0) {
		LOGL(LOG_SSENGINE, "No Attributes to SET\n");
		return S_SS_SUCCESS;	// Delta with ONLY deletes
	}
	char *item_data = SS_Malloc(item_size + 1);
	if (item_data == NULL) {
		LOGE("SS_Malloc failed!! - item_data\n");
		SS_SetUpgradeState(E_SS_MALLOC_ERROR);
		return E_SS_FAILURE;
	}
	int read_data =
		tar_get_cfg_data(ua_dataSS->update_data->ua_delta_path, ua_dataSS->update_delta->ua_attrib_path, item_data,
						 item_size);
	if (read_data <= 0) {
		LOGE("read_data failed!!\n");
		SS_SetUpgradeState(E_SS_FSBADDELTA);
		if (item_data != NULL) {
			SS_Free(item_data);
			item_data = NULL;
		}
		return E_SS_FAILURE;
	}
	pline = strtok_r(item_data, "\n", &psaveptr);
	if (pline == NULL) {
		LOGL(LOG_SSENGINE, "No Attributes to SET as no lines in file\n");
		if (item_data != NULL) {
			SS_Free(item_data);
			item_data = NULL;
		}
		return E_SS_FAILURE;
	}

	while (pline) {
		char *saveptr_pline = NULL;
		pfilePath = strtok_r(pline, "\"", &saveptr_pline);

		if (pfilePath && strcmp(pfilePath, SS_FWSLASH) == 0) {
			LOGE("\n skip root: it is RO");
			pline = strtok_r(NULL, SS_TOKEN_NEWLINE, &psaveptr);
			continue;
		}

		pfiletype = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr_pline);
		attributSize = strtok_r(NULL, SS_TOKEN_SPACE, &saveptr_pline);
		pattribs = strtok_r(NULL, SS_TOKEN_NEWLINE, &saveptr_pline);
		//LOG("\nSS_FSSetAttributes [%s][%s][%s]", pfiletype, attributSize, pattribs);
		if (pattribs && pfilePath && pfiletype) {
			ulAttribSize = strlen(pattribs);
			//LOG("\nSS_SetFileAttributes [%s][%s][%d][%s]",pfilePath,pfiletype,ulAttribSize, pattribs );
			//LOG("SS_SetFileAttributes [%s]\n", pfilePath);

			result = SS_SetFileAttributes(pfilePath, ulAttribSize, (const unsigned char *)pattribs);
			if (result != S_SS_SUCCESS) {
				LOGE("\n Failed to set Attributes %s", pfilePath);
				SS_SetUpgradeState(E_SS_FSBADATTRIBUTES);
				if (item_data) {
					SS_Free(item_data);
					item_data = NULL;
				}
				fail_cnt++;
			}
		} else {
			LOGE("\n Failed to Parse Attributes - LINE %s", pline);
			SS_SetUpgradeState(E_SS_FSBADATTRIBUTES);
			if (item_data) {
				SS_Free(item_data);
				item_data = NULL;
			}
			fail_cnt++;
		}
		pline = strtok_r(NULL, SS_TOKEN_NEWLINE, &psaveptr);
	}

	if (item_data != NULL) {
		SS_Free(item_data);
		item_data = NULL;
	}

	if (fail_cnt > 0)
		result = E_SS_FAILURE;
	else
		result = S_SS_SUCCESS;

	return result;
}

/*!
 *********************************************************************************
 *					 SS_FSUpdateFile
 *********************************************************************************
 *
 * @brief
 *	This is used to update individual files on case basis
 *
 *
 *	@param
 *
 *	@return				returns S_SS_SUCCESS
 *						E_SS_FAILURE in case of error
 *
 *********************************************************************************
 */
int SS_FSUpdateFile(int ubFileType, ua_dataSS_t * ua_dataSS, int ulPatchCount, fs_params * pFsNode,
					const char *patch_path)
{
	int ulFileIndex = 1;
	char ubPatch[SS_MAX_FILE_PATH] = { 0, };
	int ulReadCnt = 0;
	int ulResult = S_SS_SUCCESS;
	FileInfo source_file;
	FileInfo target_file;
	uint8_t target_sha1[SHA_DIGEST_SIZE] = { 0, };

	if (!patch_path) {
		LOGE("Bad patch_path name\n");
		return E_SS_FAILURE;
	}
	switch (ubFileType) {
	case DIFFS:
		{
			LOGL(LOG_SSENGINE, "DIFFS mode start\n");
#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE, "DIFFS mode start time = [%lf]\n",
			     t1);
#endif
			tar_open(ua_dataSS->update_data->ua_delta_path);
			while (pFsNode) {
				//LOGL(LOG_SSENGINE, "DIFFS update Index: [%d] \n", ulFileIndex++);
				snprintf(ubPatch, SS_MAX_FILE_PATH, "%s%s", patch_path, pFsNode->patch_name);
				//LOGL(LOG_SSENGINE, "DIFF list --- [File Name %s]\n [Patch Name %s]",pFsNode->file_path, ubPatch);
				if (pFsNode->data_size > 0) {
					ulReadCnt =
						fast_tar_extract_file(ua_dataSS->update_data->ua_delta_path, ubPatch, SS_PATCHFILE_SOURCE,
											  pFsNode->data_size, pFsNode->data_offset);
					if (ulReadCnt < 0) {
						ulResult = E_SS_FAILURE;
						tar_close();
						SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
						LOGE("Delta Read Failed\n");
						break;
					}
					//LOGL(LOG_SSENGINE,"Updating [Item - %s]and size is[%d] Read Count[%d]\n",ubPatch, pFsNode->data_size, ulReadCnt);

					if (ulReadCnt > 0)
						ulResult =
							SS_UpdateDeltaFS(pFsNode->file_old_path, pFsNode->file_new_path, pFsNode->sha1src,
											 pFsNode->sha1trg, pFsNode->data_size);
					if (ulResult != S_SS_SUCCESS) {
						LOGE("FS update Failed Result : [%d], [Item - %s]and size is[%d] Read Count[%d], index = [%d]\n", ulResult,
							 ubPatch, pFsNode->data_size, ulReadCnt, ulFileIndex);
						tar_close();
						SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
						break;
					}
					SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
					pFsNode = pFsNode->nextnode;
				}
				else {
					ulResult = E_SS_FAILURE;
					tar_close();
					LOGE("size is invalid, index = [%d]\n",
					     ulFileIndex);
					SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
					break;
				}
				ulFileIndex++;
			}
			LOGL(LOG_SSENGINE, "DIFFS : Total index = [%d]\n",
			     ulFileIndex - 1);
#ifdef TIME_PROFILING
			get_time_stamp1();  //total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "DIFFS mode end time = [%lf], diff = [%lf]\n", t2,
			     (t2 - t1));
#endif
			tar_close();
			LOGL(LOG_SSENGINE, "DIFFS mode end\n");
		}
		break;
	case MOVES:
		{
			LOGL(LOG_SSENGINE, "MOVES mode start\n");

#ifdef TIME_PROFILING
			get_time_stamp1();  //total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE, "MOVES mode start time = [%lf]\n",
			     t1);
#endif
			while (pFsNode) {
				//LOGL(LOG_SSENGINE, "MOVES update Index: [%d] \n", ulFileIndex++);
				int skip_flag = 0;
				if (SS_LoadFile(pFsNode->file_old_path, &source_file) == 0) {
					LOGL(LOG_SSENGINE, "Patch Can be applied\n");
					if (source_file.data)
						SS_Free(source_file.data);
				} else if (SS_LoadFile(pFsNode->file_new_path, &target_file) == 0) {
					LOGL(LOG_SSENGINE, "source deleted!!, file_old_path: [%s]\n", pFsNode->file_old_path);
					if (ParseSha1(pFsNode->sha1src, target_sha1) != 0) {
						LOGE("failed to parse sha1 \"%s\"\n", pFsNode->sha1src);
						ulResult = E_SS_FAILURE;
						SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
						if (target_file.data)
							SS_Free(target_file.data);
						break;
					}
					if (memcmp(target_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
						LOGL(LOG_SSENGINE, "Patch already applied\n");
						skip_flag = 1;
						if (target_file.data)
							SS_Free(target_file.data);
					} else {
						LOGL(LOG_SSENGINE, "target_sha1 diff!!: [%s]\n", target_sha1);
						ulResult = E_SS_FAILURE;
						SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
						if (target_file.data)
							SS_Free(target_file.data);
						break;
					}
				} else {
					LOGE("No exist files. - file_old_path: [%s] file_new_path: [%s]\n", pFsNode->file_old_path, pFsNode->file_new_path);
					ulResult = E_SS_FAILURE;
					SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
					break;
				}

				if (skip_flag == 0) {
					ulResult = SS_MoveFile(pFsNode->file_old_path, pFsNode->file_new_path);
					if (ulResult != S_SS_SUCCESS) {
					LOGE("Move Failed for [%s] to [%s], result = [%d], index = [%d]\n", pFsNode->file_old_path, pFsNode->file_new_path, ulResult, ulFileIndex);
						SS_SetUpgradeState(ulResult);
						break;
					} else {
						// Verification
						if (SS_LoadFile(pFsNode->file_new_path, &target_file) == 0) {
							if (ParseSha1(pFsNode->sha1src, target_sha1) != 0) {
								LOGE("failed to parse sha1 \"%s\"\n", pFsNode->sha1src);
								ulResult = E_SS_FAILURE;
								SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
								if (target_file.data)
									SS_Free(target_file.data);
								break;
							}
							if (memcmp(target_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
								LOGL(LOG_SSENGINE, "Patch success!!\n");
								if (target_file.data)
									SS_Free(target_file.data);
							} else {
								LOGL(LOG_SSENGINE, "target_sha1 diff!!: [%s]\n", target_sha1);
								ulResult = E_SS_FAILURE;
								SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
								if (target_file.data)
									SS_Free(target_file.data);
								break;
							}
						}
					}
				}
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
				pFsNode = pFsNode->nextnode;
				ulFileIndex++;
			}
			LOGL(LOG_SSENGINE, "MOVES : Total index = [%d]\n",
			     ulFileIndex - 1);
#ifdef TIME_PROFILING
			get_time_stamp1();  //total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "MOVES mode end time = [%lf], diff = [%lf]\n", t2,
			     (t2 - t1));
#endif
			LOGL(LOG_SSENGINE, "MOVES mode end\n");
		}
		break;
	case DELETES:
		{
			LOGL(LOG_SSENGINE, "DELETES mode start\n");

			int ulFiletype = 0;
#ifdef TIME_PROFILING
			get_time_stamp1();  //total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE, "DELETES mode start time = [%lf]\n",
			     t1);
#endif
			while (pFsNode) {
				if (pFsNode->type == DELETES) {
					//LOGL(LOG_SSENGINE, "DELETES update Index: [%d] \n", ulFileIndex++);
					SS_GetFileType(pFsNode->file_old_path, (enumFileType *) & ulFiletype);
					if (ulFiletype == 2)		//FT_FOLDER
						ulResult = SS_DeleteFolder(pFsNode->file_old_path);
					else
						ulResult = SS_DeleteFile(pFsNode->file_old_path);
					if (ulResult != S_SS_SUCCESS) {
						LOGE("Delete Failed, result = [%d], index = [%d]\n", ulResult, ulFileIndex);
						SS_SetUpgradeState(ulResult);
						break;
					}
					ulFileIndex++;
				}
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
				pFsNode = pFsNode->nextnode;
			}
			LOGL(LOG_SSENGINE, "DELETES : Total index = [%d]\n",
			     ulFileIndex - 1);
#ifdef TIME_PROFILING
			get_time_stamp1();  //total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "DELETES mode end time = [%lf], diff = [%lf]\n",
			     t2, (t2 - t1));
#endif
			LOGL(LOG_SSENGINE, "DELETES mode end\n");
		}
		break;
	case DELETE_END:
		{
			LOGL(LOG_SSENGINE, "DELETE_END mode start\n");

			int ulFiletype = 0;
#ifdef TIME_PROFILING
			get_time_stamp1();  //total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "DELETE_END mode start time = [%lf]\n", t1);
#endif
			while (pFsNode) {
				if (pFsNode->type == DELETE_END) {
					LOGL(LOG_SSENGINE, "DELETE_END update Index: [%d] \n", ulFileIndex++);
					SS_GetFileType(pFsNode->file_old_path, (enumFileType *) & ulFiletype);
					if (ulFiletype == 2)		//FT_FOLDER
						ulResult = SS_DeleteFolder(pFsNode->file_old_path);
					else
						ulResult = SS_DeleteFile(pFsNode->file_old_path);
					if (ulResult != S_SS_SUCCESS) {
						LOGE("Delete Failed, result = [%d], index = [%d]\n", ulResult, ulFileIndex);
						SS_SetUpgradeState(ulResult);
						break;
					}
					ulFileIndex++;
				}
				pFsNode = pFsNode->nextnode;
			}
			LOGL(LOG_SSENGINE, "DELETE_END : Total index = [%d]\n",
			     ulFileIndex - 1);
#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "DELETE_END mode start time = [%lf], diff = [%lf]\n",
			     t2, (t2 - t1));
#endif
			LOGL(LOG_SSENGINE, "DELETE_END mode end\n");
		}
		break;

	case NEWFILES:
		{
			LOGL(LOG_SSENGINE, "NEWFILES mode start\n");

#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE, "NEWFILES mode start time = [%lf]\n",
			     t1);
#endif
			LOGL(LOG_SSENGINE, "Starting New file upgrade for   [%s]\n", patch_path);
			if (tar_extract_file(ua_dataSS->update_data->ua_delta_path, (char *)patch_path, SS_NEW_COMPRESSED_FILE) >=
				0)
				if (_7zdecompress(SS_NEW_COMPRESSED_FILE) == 0)
					LOGL(LOG_SSENGINE, "7zip extracted  successfully %s\n", ua_dataSS->parti_info->ua_parti_name);
				else
					LOGL(LOG_SSENGINE, "7zip extraction error for %s\n", ua_dataSS->parti_info->ua_parti_name);
			else
				LOGL(LOG_SSENGINE, "tar extraction error for %s\n", ua_dataSS->parti_info->ua_parti_name);
			SS_DeleteFile(SS_NEW_COMPRESSED_FILE);
#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "NEWFILES mode end time = [%lf], diff = [%lf]\n",
			     t2, (t2 - t1));
#endif
			LOGL(LOG_SSENGINE, "NEWFILES mode end\n");
		}
		break;
	case SYMDIFFS:
		{
			LOGL(LOG_SSENGINE, "SYMDIFFS mode start\n");

#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE, "SYMDIFFS mode start time = [%lf]\n",
			     t1);
#endif
			while (pFsNode) {
				//LOGL(LOG_SSENGINE, "SYMDIFFS update Index: [%d] \n", ulFileIndex++);
				//LOG("Sym Diff file paths: [Linkname - %s] [reference file name- %s][]\n", pFsNode->file_path,pFsNode->patch_name);
				ulResult = SS_Unlink(pFsNode->file_old_path);
				if (ulResult == S_SS_SUCCESS) {
					ulResult = SS_Link(NULL, pFsNode->file_new_path, pFsNode->patch_name);
					if (ulResult != S_SS_SUCCESS) {
						LOGE("SS_Link Failed, Linkname:[%s], reference file name, index = [%d]:[%s]\n",
						pFsNode->file_new_path, pFsNode->patch_name, ulFileIndex);
					}
				} else {
					LOGE("Unlink Failed, result = [%d], index = [%d]\n", ulResult, ulFileIndex);
					SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
					break;
				}
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
				pFsNode = pFsNode->nextnode;
				ulFileIndex++;
			}
			LOGL(LOG_SSENGINE, "SYMDIFFS : Total index = [%d]\n",
			     ulFileIndex - 1);
#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "SYMDIFFS mode end time = [%lf], diff = [%lf]\n",
			     t2, (t2 - t1));
#endif
			LOGL(LOG_SSENGINE, "SYMDIFFS mode end\n");
		}
		break;
	case SYMNEWFILES:
		{
			LOGL(LOG_SSENGINE, "SYMNEWFILES mode start\n");

			fs_params *head_node;
			int retry_count = 0, do_retry = 0;
#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t1 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "SYMNEWFILES mode start time = [%lf]\n", t1);
#endif
 SYMLINK_CREATE:
			head_node = pFsNode;
			while (head_node) {
				//LOGL(LOG_SSENGINE, "SYMNEWS update Index: [%d] \n", ulFileIndex++);
				snprintf(ubPatch, SS_MAX_FILE_PATH, "%s%s%s", patch_path, "/", head_node->patch_name);
				LOGL(LOG_SSENGINE, "Sym New file paths: [Linkname - %s] [reference file name- %s][]\n",
					 head_node->file_old_path, head_node->patch_name);
				ulResult = SS_Link(NULL, head_node->file_old_path, head_node->patch_name);
				if (ulResult == E_SS_FAILURE) {
					LOGE("Link Failed, result = [%d], index = [%d]\n", ulResult, ulFileIndex);
					SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
					break;
				} else if (ulResult == ENOENT) {  //to handle cases where new symlink points to a new symlink yet to be created
					do_retry = 1;	   //we will retry the failed symlinks with error 2 (no file or dir) again after this cycle
					//SS_UpdateUIProgress(ua_dataSS,ulPatchCount);
					head_node = head_node->nextnode;
					continue;
				}
				SS_UpdateUIProgress(ua_dataSS, ulPatchCount, 0);
				head_node = head_node->nextnode;
				ulFileIndex++;
			}
			LOGL(LOG_SSENGINE, "SYMNEWFILES : Total index = [%d]\n",
			     ulFileIndex - 1);
			if (do_retry && (retry_count < 4)) {
				retry_count++;
				ulFileIndex = 0;
				do_retry = 0;
				goto SYMLINK_CREATE;
			} else if (do_retry && (retry_count >= 4)) {  //retry to be done maximum 4 times
				LOGE("Link Failed after %d retrys\n", retry_count);
				//SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
				break;
			}
#ifdef TIME_PROFILING
			get_time_stamp1();	//total time capturing
			t2 = atof(ts1);
			LOGL(LOG_SSENGINE,
			     "SYMNEWFILES mode end time = [%lf], diff = [%lf]\n",
			     t2, (t2 - t1));
#endif
			LOGL(LOG_SSENGINE, "SYMNEWFILES mode end\n");
		}
		break;
	default:
		break;
	}
	return ulResult;
}

#ifdef MEM_PROFILING
extern int max_mem;
extern int cur_mem;
#endif
/*!
 *********************************************************************************
 *					 SS_FSUpdatemain
 *********************************************************************************
 *
 * @brief
 *	This is the API exposed from the engine to update FS.
 *
 *
 *	@param
 *
 *	@return				returns S_SS_SUCCESS
 *						E_SS_FAILURE in case of error
 *
 *********************************************************************************
 */

int SS_FSUpdatemain(ua_dataSS_t * ua_dataSS, int part_idx)
{
	int ulResult = S_SS_SUCCESS;
	int last_update_status = 0;
	int del_type = 0;
	fs_list *head_ptr_node = NULL;
	char new_patch_path[SS_MAX_FILE_PATH] = {
		0
	};

	if (!ua_dataSS)
		return E_SS_BAD_PARAMS; // Set error ??
	head_ptr_node = headptr_list[part_idx];

	if (!head_ptr_node) {	   //in case of power failure, try rebilding nodes again
		SS_FSVerifyPartition(ua_dataSS, part_idx);
		head_ptr_node = headptr_list[part_idx];
	}

	if (!head_ptr_node)
		return E_SS_FSBADNODES;

	SS_GetPartition_LocDetails(ua_dataSS);

	LOGL(LOG_SSENGINE, "FS Update Entry PartIndex: [%d]\n", part_idx);

#ifdef POWER_FAIL_TEST
	int fail_test_flag = 0;
	fail_test_flag = SS_Get_power_fail_flag();
	LOGL(LOG_SSENGINE, "fail_test_flag: [%d]\n", fail_test_flag);
#endif

	if(SS_Get_last_update_status(&last_update_status, &del_type) == -1)
		LOGE("SS_Get_last_update_status failed!!\n");

	LOGL(LOG_SSENGINE, "last_update_status: [%d], del_type: [%d]\n", last_update_status, del_type);

	if (head_ptr_node->del_ref == NULL) {
		LOGL(LOG_SSENGINE, "No DEL header\n");
	} else if (ulResult == S_SS_SUCCESS) {
		if (del_type < DELETES) {
			ulResult = SS_FSUpdateFile(DELETES, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->del_ref,
					ua_dataSS->update_delta->ua_patch_path);
			if (ulResult == S_SS_SUCCESS) {
				LOGL(LOG_SSENGINE, "SS_FSUpdateFile - DELETES success!!\n");

#ifdef POWER_FAIL_TEST
				if (fail_test_flag < DELETES) {
					if (SS_Do_Power_fail_test(DELETES) == E_SS_FAILURE) {
						LOGE("SS_Do_Power_fail_test failed!!\n");
					}
				}
#endif
				SS_Set_last_update_status(part_idx, DELETES);
			}
		} else {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - DELETES already applied!!\n");
		}
	}

	if (head_ptr_node->dif_ref == NULL) {
		LOGL(LOG_SSENGINE, "No DIFF header\n");
	} else if (ulResult == S_SS_SUCCESS) {
		ulResult =
			SS_FSUpdateFile(DIFFS, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->dif_ref,
							ua_dataSS->update_delta->ua_patch_path);
		if (ulResult == S_SS_SUCCESS) {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - DIFFS success!!\n");
#ifdef POWER_FAIL_TEST
			if (fail_test_flag < DIFFS) {
				if (SS_Do_Power_fail_test(DIFFS) == E_SS_FAILURE) {
					LOGE("SS_Do_Power_fail_test failed!!\n");
				}
			}
#endif
		}
	}

	if (head_ptr_node->move_ref == NULL) {
		LOGL(LOG_SSENGINE, "No MOVE header\n");
	} else if (ulResult == S_SS_SUCCESS) {
		if (del_type < MOVES) {
		ulResult =
			SS_FSUpdateFile(MOVES, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->move_ref,
							ua_dataSS->update_delta->ua_patch_path);
			if (ulResult == S_SS_SUCCESS) {
				LOGL(LOG_SSENGINE, "SS_FSUpdateFile - MOVES success!!\n");
#ifdef POWER_FAIL_TEST
				if (fail_test_flag < MOVES) {
					if (SS_Do_Power_fail_test(MOVES) == E_SS_FAILURE) {
						LOGE("SS_Do_Power_fail_test failed!!\n");
					}
				}
#endif
				SS_Set_last_update_status(part_idx, MOVES);
			}
		} else {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - MOVES already applied!!\n");
		}
	}

	if (head_ptr_node->del_ref == NULL) {
		LOGL(LOG_SSENGINE, "No DEL header\n");
	} else if (ulResult == S_SS_SUCCESS) {
		if (del_type < DELETE_END) {
			ulResult = SS_FSUpdateFile(DELETE_END, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->del_ref,
					ua_dataSS->update_delta->ua_patch_path);
			if (ulResult == S_SS_SUCCESS) {
				LOGL(LOG_SSENGINE, "SS_FSUpdateFile - DELETE_END success!!\n");
#ifdef POWER_FAIL_TEST
				if (fail_test_flag < DELETE_END) {
					if (SS_Do_Power_fail_test(DELETE_END) == E_SS_FAILURE) {
						LOGE("SS_Do_Power_fail_test failed!!\n");
					}
				}
#endif
				SS_Set_last_update_status(part_idx, DELETE_END);
			}
		} else {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - DELETE_END already applied!!\n");
		}
	}

	 if (ulResult == S_SS_SUCCESS) {
		 //new file extraction start
		snprintf(new_patch_path, SS_MAX_FILE_PATH, "%s%s", ua_dataSS->parti_info->ua_subject_name, SS_COMPRESSED_FILE);	 // subject name wil have fw slash as part of cfg file
		LOGL(LOG_SSENGINE, "File path created to extract new files : [%s]\n", new_patch_path);
		ulResult =
			SS_FSUpdateFile(NEWFILES, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->new_ref, new_patch_path);
		//new file extraction end
		if (ulResult == S_SS_SUCCESS) {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - NEWFILES success!!\n");
		}
	}

	if (head_ptr_node->sym_difref == NULL) {
		LOGL(LOG_SSENGINE, "No SYMDIFF header\n");
	} else if (ulResult == S_SS_SUCCESS) {
		ulResult =
			SS_FSUpdateFile(SYMDIFFS, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->sym_difref,
							ua_dataSS->update_delta->ua_patch_path);
		if (ulResult == S_SS_SUCCESS) {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - SYMDIFFS success!!\n");
#ifdef POWER_FAIL_TEST
			if (fail_test_flag < SYMDIFFS) {
				if (SS_Do_Power_fail_test(SYMDIFFS) == E_SS_FAILURE) {
					LOGE("SS_Do_Power_fail_test failed!!\n");
				}
			}
#endif
		}
	}

	if (head_ptr_node->sym_newref == NULL) {
		LOGL(LOG_SSENGINE, "No SYMNEW header\n");
	} else if (ulResult == S_SS_SUCCESS) {
		ulResult =
			SS_FSUpdateFile(SYMNEWFILES, ua_dataSS, head_ptr_node->ulPatchCount, head_ptr_node->sym_newref,
							ua_dataSS->update_delta->ua_patch_path);
		if (ulResult == S_SS_SUCCESS) {
			LOGL(LOG_SSENGINE, "SS_FSUpdateFile - SYMNEWFILES success!!\n");
#ifdef POWER_FAIL_TEST
			if (fail_test_flag < SYMNEWFILES) {
				if (SS_Do_Power_fail_test(SYMNEWFILES) == E_SS_FAILURE) {
					LOGE("SS_Do_Power_fail_test failed!!\n");
				}
			}
#endif
		}
	}


	if (ulResult == S_SS_SUCCESS) {
		ulResult = SS_FSSetAttributes(ua_dataSS);
	} else {
		SS_FSSetAttributes(ua_dataSS);	// To prevent boot failures by smack.
	}

	sync();
	sleep(1);
	SS_FSClearNodes(part_idx);

	if (ulResult == S_SS_SUCCESS) {
		SS_UpdateUIProgress(ua_dataSS, 0, 1);	//fix WGID : 51963, When all updates are done to FS , patchcount is not needed, passing 1 to 3rd arg is enough
		SS_Set_last_update_status(part_idx, DEL_TYPE_MAX);
	}

	LOGL(LOG_SSENGINE, "FS update Complete PartIndex: [%d]\n", part_idx);
#ifdef  MEM_PROFILING
	LOGL(LOG_SSENGINE, "Stats are : Cur Max : [%d] Global Max : [%d]\n", cur_mem, max_mem);
#endif
#ifdef POWER_FAIL_TEST
	unlink(SS_POWER_FAIL_TEST_FLAG);
#endif
	if (ulResult == S_SS_SUCCESS)
		return ulResult;
	else
		return SS_GetUpgradeState();
}

/*!
 *********************************************************************************
 *					 SS_FSUpdatemain
 *********************************************************************************
 *
 * @brief
 *	This is the API exposed from the engine to update FS.
 *	FS entry function for updating FS partition. Should be invoked only after verification of the partition
 *
 *
 *	@param				Requires common data structure having all details & Partition Index.
 *						(Used for getting right NODES information that built during verification)
 *						(Configuration, Delta info, Partition Info, UI link , Kind of operation.(Verify or Updates))
 *
 *	@return				returns S_SS_SUCCESS
 *						E_SS_FAILURE in case of error
 *
 *********************************************************************************
 */
size_t SS_FSAvailiableFreeSpace(char *block_name)
{

	struct mntent *ent;
	FILE *aFile;
	struct statfs sb;
	aFile = setmntent("/proc/mounts", "r");
	if (aFile == NULL) {
		LOGE("setmntent error");
		return E_SS_FAILURE;
	}
	while (NULL != (ent = getmntent(aFile))) {
		if (strcmp(ent->mnt_fsname, block_name) == 0) {
			if (statfs(ent->mnt_dir, &sb) == 0)
				LOGL(LOG_SSENGINE, "Total free space = %ld, blocks free = %ld\n", sb.f_bsize * sb.f_bavail, sb.f_bfree);
		}
	}
	endmntent(aFile);
	return ((long long)sb.f_bsize * (long long)sb.f_bavail >= (long long)SIZE_4GB) ? SIZE_4GB : sb.f_bsize * sb.f_bavail ;
}

int SS_FSVerifyPartition(ua_dataSS_t * ua_dataSS, int part_idx)
{
	int ulResult = S_SS_SUCCESS;
	size_t free_space = 0;
	if (!ua_dataSS) {
		LOGE("Wrong Param for fs verification\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}

	LOGL(LOG_SSENGINE, "FS max free mem reqired : [%d]\n", ua_dataSS->update_cfg->soure_img_size);
	free_space = SS_FSAvailiableFreeSpace(ua_dataSS->parti_info->ua_blk_name);
	if (free_space != E_SS_FAILURE) {
		//Source img size is max single file size for a file system under upgrade, which is updated in CFG file by UPG
		if ((free_space) < (ua_dataSS->update_cfg->soure_img_size + ua_dataSS->update_cfg->soure_img_size / 10)) {
			LOGE("Not enough free space [%d] for max size [%d]\n", free_space,
				 (ua_dataSS->update_cfg->soure_img_size + ua_dataSS->update_cfg->soure_img_size / 10));
			//SS_SetUpgradeState(E_SS_FSMEMORYERROR);
			//return E_SS_FAILURE;
		} else
			LOGL(LOG_SSENGINE, "Enough space for Partition [%s]\n", ua_dataSS->parti_info->ua_parti_name);
	}

	SS_GetAvailableFreeSpace(SS_COMMON_WORKSPACE, &free_space);
	//Checking for 2 times the file size free space , as delta can be worst case size of file.
	if ((free_space) < (2 * ua_dataSS->update_cfg->soure_img_size)) {
		LOGE("Not enough free space [%d] for max size [%d]\n", free_space, (2 * ua_dataSS->update_cfg->soure_img_size));
		SS_SetUpgradeState(E_SS_FSMEMORYERROR);
		return E_SS_FAILURE;
	}
#ifdef MEM_PROFILING
	if (!mem_profiling_start)
		if (!(S_SS_SUCCESS == SS_Do_Memory_Profiling()))
			return E_SS_FAILURE;
#endif
	SS_GetPartition_LocDetails(ua_dataSS);
	LOGL(LOG_SSENGINE, "FS Verification Start PartIndex:[%d]\n", part_idx);
	if (ua_dataSS->ua_operation == UI_OP_SCOUT)
		gvalid_session = 1;  // (shd b true if called during verification)
	headptr_list[part_idx] = SS_FSBuildNodes(ua_dataSS);
#ifdef TIME_PROFILING
	LOGL(LOG_SSENGINE, "fast_tar_get_item_size_time :[%lf]\n", fast_tar_get_item_size_time);
	LOGL(LOG_SSENGINE, "SS_LoadFile_time :[%lf]\n", SS_LoadFile_time);
	LOGL(LOG_SSENGINE, "SS_FSBuildNodes_time :[%lf]\n", SS_FSBuildNodes_time);
#endif
	if (!headptr_list[part_idx]) {
		LOGE("FS Verification Failed PartIndex: [%d]\n", part_idx);
		SS_FSClearNodes(part_idx);
		ulResult = E_SS_FAILURE;
	}

	if (ulResult == S_SS_SUCCESS)
		return ulResult;
	else
		return SS_GetUpgradeState();
}

//Should check if space is available????
int SS_BackupSource(const char *source_filename)
{
	int ret = E_SS_FAILURE;

	if (source_filename) {
		ret = (int)SS_CopyFile(source_filename, SS_BACKUP_SOURCE);
		if (ret != S_SS_SUCCESS) {
			LOGE("failed to back up source file  Error [%d]\n", ret);
			SS_SetUpgradeState(E_SS_FSSRCBACKUPFAILED);
		}
	}
	return ret;
}

int SS_BackupSourceClear(void)
{
	int ret = E_SS_FAILURE;
	ret = (int)SS_DeleteFile(SS_BACKUP_SOURCE);
	if (ret != S_SS_SUCCESS) {
		LOGE("failed to delete BACKUP file\n");
		SS_SetUpgradeState(E_SS_FSSRCBACKUPFAILED);
	}
	return ret;
}

int SS_PatchSourceClear(void)
{
	int ret = E_SS_FAILURE;
	ret = (int)SS_DeleteFile(SS_PATCHFILE_SOURCE);
	if (ret != S_SS_SUCCESS) {
		LOGE("failed to delete PATCHFILE file\n");
		SS_SetUpgradeState(E_SS_PATCHFILE_DEL_ERROR);
	}
	return ret;
}
int SS_IMGVerifyFullImage(ua_dataSS_t * ua_dataSS)
{
	int read_cnt = 0, patch_size = 0;
	FileInfo source_file;
	uint8_t target_sha1[SHA_DIGEST_SIZE];
	int ulResult = S_SS_SUCCESS;

	if (!(ua_dataSS && ua_dataSS->update_cfg && ua_dataSS->parti_info &&
				ua_dataSS->parti_info->ua_blk_name &&
				!(ua_dataSS->update_cfg->skip_verify == 1))) {
		LOGE("Bad structure or members\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}

	LOGL(LOG_SSENGINE, "FULL IMG Verification Entry BlkName:[%s]\n", ua_dataSS->parti_info->ua_blk_name);

	if (ua_dataSS->update_data && ua_dataSS->parti_info && ua_dataSS->update_data->ua_delta_path
		&& ua_dataSS->parti_info->ua_subject_name) {
		patch_size = tar_get_item_size(ua_dataSS->update_data->ua_delta_path, ua_dataSS->parti_info->ua_subject_name);
	} else {
		LOGE("Bad structure members in ua_dataSS\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	if (ua_dataSS->update_cfg && ua_dataSS->update_cfg->soure_img_size && ua_dataSS->update_cfg->target_sha1) {
		LOGL(LOG_SSENGINE, "\nParams -image size [%d] sha1 [%s]\n",
			 ua_dataSS->update_cfg->soure_img_size, ua_dataSS->update_cfg->target_sha1);
	} else {
		LOGE("Bad structure member update_cfg in ua_dataSS\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	if (ParseSha1(ua_dataSS->update_cfg->target_sha1, target_sha1) != 0) {
		LOGE("failed to parse tgt-sha1 \"%s\"\n", ua_dataSS->update_cfg->target_sha1);
		SS_SetUpgradeState(E_SS_IMGBADDELTA);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}

	if ((patch_size) > 0)
		read_cnt = tar_extract_file(ua_dataSS->update_data->ua_delta_path,
				ua_dataSS->parti_info->ua_subject_name, SS_PATCHFILE_SOURCE);

	if (read_cnt <= 0) {
		LOGL(LOG_SSENGINE, "Failed to read delta\n");
		SS_SetUpgradeState(E_SS_IMGBADDELTA);
		ulResult = E_SS_FAILURE;
		goto Cleanup;
	}
	if (SS_LoadFile(SS_PATCHFILE_SOURCE, &source_file) == 0) {
		if (memcmp(source_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
			LOGL(LOG_SSENGINE, "Patch Can be applied\n");
			SS_Free(source_file.data);
			ulResult = S_SS_SUCCESS;
		} else{
			LOGL(LOG_SSENGINE, "Patch Cannot be applied\n");
			SS_Free(source_file.data);
			SS_SetUpgradeState(E_SS_IMGBADDELTA);
			ulResult = E_SS_FAILURE;
			goto Cleanup;
		}
	} else {
		LOGL(LOG_SSENGINE, "Failed to LoadFile\n");
		SS_SetUpgradeState(E_SS_IMGBADDELTA);
		ulResult = E_SS_FAILURE;
	}

Cleanup:
	if (file_exist(SS_PATCHFILE_SOURCE))
		SS_DeleteFile(SS_PATCHFILE_SOURCE);
	return ulResult;
}

int SS_IMGVerfiyPartition(ua_dataSS_t * ua_dataSS)
{
	FileInfo source_file;
	int ulResult = S_SS_SUCCESS;
	uint8_t source_sha1[SHA_DIGEST_SIZE];
	uint8_t target_sha1[SHA_DIGEST_SIZE];
	size_t free_space = 0;

	if (!(ua_dataSS && ua_dataSS->update_cfg && ua_dataSS->parti_info && ua_dataSS->parti_info->ua_blk_name)) {
		LOGE("Bad structure or members\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}

	//We verify twice the image size for BACKUP source, not on Partition. As Patch will be created on RAM
	SS_GetAvailableFreeSpace(SS_COMMON_WORKSPACE, &free_space);
	if ((free_space) < (2 * ua_dataSS->update_cfg->target_img_size)) {
		LOGE("Not enough free space [%d] for twice max size [%d]\n", free_space,
			 (2 * ua_dataSS->update_cfg->target_img_size));
		SS_SetUpgradeState(E_SS_FSMEMORYERROR);
		return E_SS_FAILURE;
	}

	if (ParseSha1(ua_dataSS->update_cfg->soure_sha1, source_sha1) != 0) {
		LOGE("failed to parse Src-sha1 \"%s\"\n", ua_dataSS->update_cfg->soure_sha1);
		SS_SetUpgradeState(E_SS_SHAPRASE_FAILED);
		return E_SS_FAILURE;
	}
	// corner case, Parsing sha can fail if update.cfg is wrong/manually edited
	if (ParseSha1(ua_dataSS->update_cfg->target_sha1, target_sha1) != 0) {
		LOGE("failed to parse Target-sha1 \"%s\"\n", ua_dataSS->update_cfg->target_sha1);
		SS_SetUpgradeState(E_SS_SHAPRASE_FAILED);
		return E_SS_FAILURE;
	}

	source_file.size = ua_dataSS->update_cfg->soure_img_size;
	source_file.data = NULL;
	if (SS_LoadPartition(ua_dataSS->parti_info->ua_blk_name, &source_file) == 0) {
		if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
			LOGL(LOG_SSENGINE, "SS_IMGVerfiyPartition - SHA matches with source [%s] \n",
				 ua_dataSS->parti_info->ua_blk_name);
		} else {			  // Need not compare with Target sha as once upgraded, it should NOT verify same partition again.
			unsigned char actualShaBuffer[41] = { 0, };
			hex_digest(source_file.sha1, actualShaBuffer, SHA_DIGEST_SIZE);
			LOGE("SS_IMGVerfiyPartition - SHA mismatch with SRC [%s] Expected [%s] Actual [%s]\n",
				ua_dataSS->parti_info->ua_blk_name, ua_dataSS->update_cfg->soure_sha1, actualShaBuffer);
			SS_SetUpgradeState(E_SS_IMGSRCCURRUPTED);
			ulResult = E_SS_FAILURE;
		}
	}
	SS_Free(source_file.data);
	if (ulResult == S_SS_SUCCESS) {
		if (ua_dataSS->ui_progress)
			ua_dataSS->ui_progress(ua_dataSS, 100);
		return ulResult;
	} else
		return SS_GetUpgradeState();
}

/*!
 *********************************************************************************
 *					 SS_IMGUpdatemain
 *********************************************************************************
 *
 * @brief
 *	This is the API exposed from the engine to update Images(FULL and DELTA images)
 *
 *
 *	@param
 *
 *	@return				returns S_SS_SUCCESS
 *						E_SS_FAILURE in case of error
 *
 *********************************************************************************
 */

int SS_IMGUpdatemain(ua_dataSS_t * ua_dataSS, int update_type)  //SS_FSUpdatePartition
{
	int read_cnt = 0, patch_size;
	int ulResult = S_SS_SUCCESS;

	//sprintf(Blk_name, "%s%s%s","EMMC",":", ua_dataSS->parti_info->ua_blk_name);
	//LOGL(LOG_SSENGINE, "IMG Upgrade Entry BlkName:[%s]\n",Blk_name);
	if (!ua_dataSS) {
		LOGE("Bad structure ua_dataSS\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}
	LOGL(LOG_SSENGINE, "IMG Upgrade Entry BlkName:[%s]\n", ua_dataSS->parti_info->ua_blk_name);

	if (ua_dataSS->update_data && ua_dataSS->parti_info && ua_dataSS->update_data->ua_delta_path
		&& ua_dataSS->parti_info->ua_subject_name)
		patch_size = tar_get_item_size(ua_dataSS->update_data->ua_delta_path, ua_dataSS->parti_info->ua_subject_name);
	else {
		LOGE("Bad structure members in ua_dataSS\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}
	if (ua_dataSS->update_cfg && ua_dataSS->update_cfg->soure_img_size && ua_dataSS->update_cfg->target_sha1)
		LOGL(LOG_SSENGINE, "\n SS_IMGUpdatemain Params -source size [%d] sha1 [%s]",
			 ua_dataSS->update_cfg->soure_img_size, ua_dataSS->update_cfg->target_sha1);
	else {
		LOGE("Bad structure member update_cfg in ua_dataSS\n");
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}

	if ((patch_size) > 0)
		read_cnt =
			tar_extract_file(ua_dataSS->update_data->ua_delta_path, ua_dataSS->parti_info->ua_subject_name,
							 SS_PATCHFILE_SOURCE);

	if (read_cnt <= 0) {
		ulResult = E_SS_FAILURE;
		SS_SetUpgradeState(E_SS_IMGBADDELTA);
		return E_SS_FAILURE;	//ulResult;
	}
	if (ua_dataSS->ui_progress)
		ua_dataSS->ui_progress(ua_dataSS, 40);

	if (update_type == FULL_IMG && ua_dataSS->update_data->ua_temp_path)
		ulResult = SS_MoveFile(SS_PATCHFILE_SOURCE, ua_dataSS->update_data->ua_temp_path);
	else if ((ua_dataSS->update_cfg->update_type == DELTA_IMG && ua_dataSS->write_data_to_blkdev)
					|| ua_dataSS->update_cfg->update_type == EXTRA) {

		FILE *fp = NULL;
		char buf[14] = { 0, };  //to store zImage-delta magic keyword
		char err_buf[256];
		ssize_t bytes_read;
		fp = fopen(SS_PATCHFILE_SOURCE, "r");
		if (fp == NULL) {
			strerror_r(errno, err_buf, sizeof(err_buf));
			LOGE("Failed to open patch file Error:[%s]\n", err_buf);
			SS_SetUpgradeState(E_SS_FSFAILEDTOOPENPATCHINFO);
			return E_SS_FAILURE;
		}
		bytes_read = fread(buf, 1, 13, fp);		//error check not required as any delta corruption will be caught in SS_UpdateDeltaIMG
		if (bytes_read != 13)
			LOGL(LOG_SSENGINE, "short read of \"%s\" (%ld bytes of %ld)\n", SS_PATCHFILE_SOURCE, (long)bytes_read, (long)13);
		fclose(fp);

		if (strcmp(buf, SS_KERNEL_MAGIC) == 0)
			ulResult = SS_UpdateDeltaKernel(ua_dataSS, ua_dataSS->write_data_to_blkdev);
		else
			ulResult = SS_UpdateDeltaIMG(ua_dataSS, ua_dataSS->write_data_to_blkdev);
	} else {
		LOGE("Update type is INVALID - Exit \n");
		ulResult = E_SS_FAILURE;
		SS_SetUpgradeState(E_SS_BAD_PARAMS);
		return E_SS_FAILURE;
	}

	if (ulResult == S_SS_SUCCESS) {
		if (ua_dataSS->ui_progress)
			ua_dataSS->ui_progress(ua_dataSS, 100);
		return ulResult;
	} else
		return SS_GetUpgradeState();
}
