/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _SS_UPI_H_
#define _SS_UPI_H_

#define DISPLAYRESOLUTION_SIZE 50

struct details {
	int diffs;
	int moves;
	int news;
	int deletes;
	int symdiffs;
	int symnews;
};

enum DEL_TYPE { DELETES = 1, DIFFS, MOVES, DELETE_END, NEWFILES, SYMDIFFS, SYMNEWFILES, DEL_TYPE_MAX };
struct fs_params {			  // Use Macros
	char file_old_path[512];
	char file_new_path[512];
	char patch_name[256];
	char sha1src[64];
	char sha1trg[64];
	int data_size;
	int data_offset;
	int type;				   //0 is for diff and 1 is for verbatim
	struct fs_params *nextnode;
};
typedef struct fs_params fs_params;

struct fs_list {
	fs_params *dif_ref;
	fs_params *move_ref;
	fs_params *new_ref;
	fs_params *del_ref;
	fs_params *sym_difref;
	fs_params *sym_newref;
	int ulPatchCount;
};
typedef struct fs_list fs_list;
struct details *get_fs_details(char *filename);

int SS_AppendNode(const char *ubDeltaPath, fs_params ** headparam, fs_params ** tailparam, const char *old_path,
				  const char *new_path, const char *patchname, const char *sha1src, const char *sha1trg, int type,
				  char *patchpath_name);
extern int SS_IMGUpdatemain(ua_dataSS_t * ua_dataSS, int update_type);
extern int SS_IMGVerfiyPartition(ua_dataSS_t * ua_dataSS);
extern int SS_FSUpdatemain(ua_dataSS_t * ua_dataSS, int part_idx);
extern int SS_FSVerifyPartition(ua_dataSS_t * ua_dataSS, int part_idx);
extern int SS_UpdateDeltaIMG(ua_dataSS_t * ua_dataSS, int (*write_to_blkdev) (char *, int, int, char *));
extern int SS_UpdateDeltaKernel(ua_dataSS_t * ua_dataSS, int (*write_to_blkdev) (char *, int, int, char *));

//extra functions
extern void *SS_Malloc(unsigned int size);
extern int tar_get_item_size_from_struct(tar_Data_t **, const char *, int *, int *);
extern void SS_Free(void *pMemBlock);
extern int tar_get_item_size(char *tar, char *item);
extern int tar_get_cfg_data(char *tar, char *item, char *buf, int buflen);
extern tar_Data_t *tar_build_cfg_table(char *tar);
extern int tar_open(char *tar);
extern int fast_tar_extract_file(char *tar, char *item, char *pathname, int size, int offset);
extern int tar_close();
extern int SS_UpdateDeltaFS(const char *source_filename, const char *target_filename,
							const char *source_sha1_str, const char *target_sha1_str, int patch_data_size);
extern int tar_extract_file(char *tar, char *item, char *pathname);
extern int _7zdecompress(char *path);
extern void tar_free_cfg_table(tar_Data_t ** delta_tar);
extern long SS_GetFileType(char *pLinkName, enumFileType * fileType);

extern int SS_Get_last_update_status(int* last_update_status, int* del_type);
extern void SS_Set_last_update_status(int last_update_status, int del_type);

#endif						  //_SS_UPI_H_
