/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __FOTA_LOG_H__
#define __FOTA_LOG_H__

#include <stdio.h>

/*
 * DEBUGGING FEATURE
 */

extern unsigned int __log_level__;
extern FILE *__log_out_file__;
extern int log_printf(FILE* log_fp, char* format_str, ...);
extern void truncate_log_file(char *log_path, int size_kb);
extern void set_max_logfile_size(void);

#define LOG_INFO	(1<<8)
#define LOG_ENGINE	(1<<7)
#define LOG_FUNCS	(1<<6)
#define LOG_GUI		(1<<5)
#define LOG_DEBUG	(1<<4)
#define LOG_FILE	(1<<3)
#define LOG_FLASH	(1<<2)
#define LOG_SSENGINE	LOG_ENGINE

//#define DEBUG_STDOUT
#define DEBUG_FILE

#ifdef DEBUG_STDOUT
#define LOGE(s, args...) printf("UA/ERROR(%s)  " s, __func__, ##args)   // Error log
#define LOGL(mask, s, args...) do { if ((mask) & __log_level__) printf("UA/(%s): " s, __func__, ##args); } while (0)
#define LOG(s, args...) LOGL(LOG_DEBUG, s, ##args)

#elif defined(DEBUG_FILE)
#define LOGE(s, args...) (void)log_printf(__log_out_file__, "UA/ERROR(%s)  " s, __func__, ##args)
#define LOGL(mask, s, args...) do { if ((mask) & __log_level__) (void)log_printf(__log_out_file__, "UA/(%s): " s , __func__, ##args); } while (0)
#define LOG(s, args...) LOGL(LOG_DEBUG, s, ##args)

#elif defined(DEBUG_STDOUT_FILE)        // debug printf
#define LOGE(s, args...) do {\
						printf("UA/ERROR(%s)  " s, __func__, ##args);\
						(void)log_printf(__log_out_file__, "UA/ERROR(%s)  " s, __func__, ##args);\
					} while (0)
#define LOGL(mask, s, args...) do { \
						if ((mask) & __log_level__) {\
							printf("UA/(%s): " s , __func__, ##args);\
							(void)log_printf(__log_out_file__, "UA/(%s): " s, __func__, ##args);\
						} \
					} while (0)
#define LOG(s, args...) LOGL(LOG_DEBUG, s, ##args)

#else
#define LOGE(s, args...)
#define LOGL(mask, s, args...)
#define LOG(s, args...)

#endif

#endif                          /* __FOTA_LOG_H__ */
